import uuid
from datetime import datetime
from typing import TypeVar

from sqlalchemy import DateTime, String
from sqlalchemy.orm import Mapped, mapped_column

from skys_llc_auth.databases import Base

T = TypeVar("T", bound="CredentialStorage")


class CredentialStorage(Base):
    __tablename__ = "credential_storage"

    id: Mapped[uuid.UUID] = mapped_column(primary_key=True, default=uuid.uuid4)
    access_token: Mapped[str] = mapped_column(String(2500), nullable=True)
    refresh_token: Mapped[str] = mapped_column(String(2500), nullable=True)
    login: Mapped[str] = mapped_column(nullable=False)
    password: Mapped[str] = mapped_column(nullable=False)
    access_until: Mapped[datetime] = mapped_column(DateTime(timezone=True), nullable=True)
    created_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), nullable=False)
    service_name: Mapped[str] = mapped_column(nullable=True)



import asyncio
import json
from collections.abc import AsyncGenerator
from typing import Any

from aiokafka import AIOKafkaConsumer, AIOKafkaProducer, ConsumerRecord

from skys_llc_auth.utils import Singleton


class KafkaManager(Singleton):
    def __init__(
        self,
        *,
        bootstrap_servers: str,
        topics_consume: list[str],
    ):
        self.loop = asyncio.get_event_loop()
        self.topics = topics_consume
        self.producer = AIOKafkaProducer(
            bootstrap_servers=bootstrap_servers,
            value_serializer=self.serializer,
            loop=self.loop,
        )
        self.consumer = AIOKafkaConsumer(
            *self.topics,
            bootstrap_servers=bootstrap_servers,
            loop=self.loop,
            value_deserializer=self.deserializer,
        )

    async def send_one(self, topic: str, payload: dict, **kw: Any):
        await self.producer.send(topic=topic, value=payload, **kw)

    def serializer(self, value: Any) -> bytes:
        return json.dumps(value).encode()

    def deserializer(self, serialized: bytes) -> Any:
        return json.loads(serialized)

    async def consume(self) -> AsyncGenerator[ConsumerRecord, None]:
        async for msg in self.consumer:
            yield msg
            if self.consumer._closed:
                return

    async def close(self):
        await self.consumer.stop()
        await self.producer.stop()



from fastapi import HTTPException


class AuthError(Exception):
    pass


class TokenError(AuthError, HTTPException):
    """Raised when token is invalid"""

    def __init__(self, detail: str, status_code: int):
        self.detail = detail
        self.status_code = status_code


class ParamsError(AuthError):
    pass


class RequestError(AuthError):
    """Raised when request is invalid or it could not be processed"""



from skys_llc_auth.cryptography import Kuznechik, Streebog
from skys_llc_auth.exceptions import AuthError, ParamsError, TokenError
from skys_llc_auth.kafka_manager import KafkaManager
from skys_llc_auth.microservices import RequestBetweenMicroservices
from skys_llc_auth.models import CredentialStorage
from skys_llc_auth.token_validation import (
    DefaultTokenParams,
    TokenValidation,
    get_token_from_request,
)
from skys_llc_auth.utils import TokenType, UserRole

__all__ = (
    "AuthError",
    "CredentialStorage",
    "DefaultTokenParams",
    "KafkaManager",
    "Kuznechik",
    "ParamsError",
    "RequestBetweenMicroservices",
    "Streebog",
    "TokenError",
    "TokenType",
    "TokenValidation",
    "UserRole",
    "get_token_from_request",
)



from enum import Enum
from typing import Any
from uuid import UUID

from skys_llc_auth.exceptions import TokenError


class UserRole(Enum):
    Student = "Student"
    Instructor = "Instructor"
    Administrator = "Administrator"
    Server = "Server"


class TokenType(str, Enum):
    access = "access"
    refresh = "refresh"


class Singleton:  # pragma: no cover
    def __new__(cls, *args: Any, **kwds: Any):
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it
        cls.__it__ = it = object.__new__(cls)
        it.init(*args, **kwds)
        return it

    def init(self, *args: Any, **kwds: Any):
        pass


def convert_jwt_str_to_uuid(field: str) -> UUID:
    try:
        return UUID(field)
    except ValueError:
        raise TokenError(status_code=400, detail="Invalid UUID format") from None



from datetime import UTC, datetime, timedelta
from typing import Any

import httpx
from httpx._types import RequestContent, RequestFiles
from loguru._logger import Logger
from sqlalchemy import select, update
from sqlalchemy.ext.asyncio import AsyncSession

from skys_llc_auth.databases import DatabaseConfig
from skys_llc_auth.exceptions import RequestError
from skys_llc_auth.models import CredentialStorage
from skys_llc_auth.schemas import Credentails


class RequestBetweenMicroservices:
    def __init__(
        self,
        refresh_url: str,
        login_url: str,
        name: str,
        access: str,
        refresh: str,
        login: str,
        password: str,
        retries: int,
        db_config: DatabaseConfig,
        logger: Logger,
        token_lifetime: int = 1440,
    ):
        self.refresh_url = refresh_url
        self.login_url = login_url
        self.name = name
        self.access_token = access
        self.refresh_token = refresh
        self.login = login
        self.password = password
        self.transport = httpx.AsyncHTTPTransport(retries=retries)
        self.headers = {}
        self.db_config = db_config
        self.token_lifetime = token_lifetime
        self.logger = logger
        self.session = None

    async def _send_request(
        self,
        method: str,
        url: str,
        content: RequestContent | None = None,
        data: dict[str, Any] | None = None,
        files: RequestFiles | None = None,
        json: Any | None = None,
        headers: dict[str, str] | None = None,
        params: dict[str, Any] | None = None,
        timeout: float | None = None,
    ) -> httpx.Response:
        """Function for send async request to another microservices"""
        self.logger.info(f"Request with microservices token to {url} with {method} method")

        async with httpx.AsyncClient(transport=self.transport) as client:
            response = await client.request(
                method,
                url,
                content=content,
                data=data,
                files=files,
                json=json,
                params=params,
                headers=headers,
                timeout=timeout,
            )
            self.logger.info(f"Response ending with status code:{response.status_code} body:{response.text}")

        return response

    async def request_with_microservice_tokens(
        self,
        method: str,
        url: str,
        *args: Any,
        **kwargs: Any,
    ) -> httpx.Response:
        """Function for send async request to another microservices and validate credentials"""
        if not self.access_token:
            self.logger.info("Access token is not present")
            await self.logging_in()

        auth = {"Authorization": "Bearer " + self.access_token}
        self.headers = kwargs.get("headers", {})
        self.headers.update(auth)

        response = await self._send_request(method=method, url=url, headers=self.headers, *args, **kwargs)  # noqa: B026

        if response.status_code == 401:
            refreshed_token_pair = await self.refresh_tokens()

            if refreshed_token_pair.status_code == 401:
                await self.logging_in()

                self.headers.update({"Authorization": "Bearer " + self.access_token})

                return await self._send_request(method=method, url=url, headers=self.headers, *args, **kwargs)  # noqa: B026

            return await self._send_request(method=method, url=url, headers=self.headers, *args, **kwargs)  # noqa: B026

        return response

    async def logging_in(self) -> httpx.Response:
        """Function for send async request to users and get tokens"""
        cred = await self.get_credentials_from_table()
        if cred:
            self.access_token = cred.access_token
            self.refresh_token = cred.refresh_token
            self.login = cred.login
            self.password = cred.password
            self.logger.info(f"Cred {self.name} found in db")

        response = await self._send_request(
            "POST", self.login_url, json={"login": self.login, "password": self.password}
        )

        if response.status_code == 401:
            self.logger.error("An error while logging, try to get cred form db")
            cred = await self.get_credentials_from_table()
            if not cred:
                self.logger.error("An error while getting cred form db")
                raise RequestError(f"Cred {self.name} dissapeared")

            second_try = await self._send_request(
                "POST", self.login_url, json={"login": cred.login, "password": cred.password}
            )
            if second_try.status_code == 401:
                self.logger.error("An error while getting cred form db")
                raise RequestError(f"Login failed for {self.name} because refresh_token: {self.refresh_token}")

            elif second_try.status_code == 201:
                self.access_token = second_try.json().get("access_token", "")
                self.refresh_token = second_try.json().get("refresh_token", "")
                await self.validate_entity()

        elif response.status_code == 201:
            self.logger.info("Logging successfully completed")
            self.access_token = response.json().get("access_token", "")
            self.refresh_token = response.json().get("refresh_token", "")
            await self.validate_entity()

        return response

    async def refresh_tokens(self) -> httpx.Response:
        """Function for send async request to users and refresh new tokens"""
        response = await self._send_request(
            "POST",
            self.refresh_url,
            headers={"Authorization": "Bearer " + self.refresh_token},
        )
        if response.status_code == 401:
            raise RequestError(f"Login failed for {self.name} because refresh_token: {self.refresh_token}")

        if response.status_code == 201:
            self.access_token = response.json().get("access_token", "")
            self.refresh_token = response.json().get("refresh_token", "")

        return response

    async def insert_credentials_to_table(
        self,
        payload: Credentails,
        db: AsyncSession,
    ) -> CredentialStorage | None:
        stmt = CredentialStorage(**payload.model_dump())
        db.add(stmt)
        await db.commit()
        return stmt

    async def get_credentials_from_table(
        self,
        db: AsyncSession | None = None,
    ) -> CredentialStorage | None:
        query = (
            select(CredentialStorage)
            .where(CredentialStorage.service_name == self.name)
            .order_by(CredentialStorage.created_at.desc())
            .limit(1)
        )
        if db is None:
            db = await self._get_async_session()
            result = await db.execute(query)
            await db.close()
            return result.scalar()

        result = await db.execute(query)
        return result.scalar()

    async def update_credentials_to_table(
        self,
        payload: Credentails,
        db: AsyncSession,
    ):
        query = (
            update(CredentialStorage)
            .where(CredentialStorage.service_name == self.name)
            .values(**payload.model_dump(exclude_none=True))
            .returning(CredentialStorage)
        )
        result = await db.execute(query)
        await db.commit()
        return result.scalar()

    async def _get_async_session(self) -> AsyncSession:
        it = self.db_config.get_async_session()
        self.session = await it.__anext__()
        return self.session

    async def validate_entity(self):
        payload = Credentails(
            access_token=self.access_token,
            refresh_token=self.refresh_token,
            login=self.login,
            password=self.password,
            service_name=self.name,
            access_until=datetime.now(UTC) + timedelta(seconds=self.token_lifetime),
            created_at=datetime.now(UTC),
        )
        async with self.db_config.async_session_maker() as db:
            if await self.get_credentials_from_table(db=db):
                self.logger.info(f"Credentials for {self.name} already exist")
                await self.update_credentials_to_table(payload=payload, db=db)
                self.logger.info(f"Credentials for {self.name} updated")

            else:
                await self.insert_credentials_to_table(payload=payload, db=db)
                self.logger.info(f"Credentials for {self.name} inserting")



from collections.abc import AsyncGenerator

from sqlalchemy.ext.asyncio import (
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.orm import DeclarativeBase


class Base(DeclarativeBase):
    pass


class DatabaseConfig:
    def __init__(
        self,
        db_url_postgresql: str,
    ):
        self.db_url_postgresql = db_url_postgresql

    @property
    def engine(self):
        return create_async_engine(self.db_url_postgresql, echo=True)

    @property
    def async_session_maker(self):
        return async_sessionmaker(
            self.engine,
            class_=AsyncSession,
            expire_on_commit=False,
        )

    async def get_async_session(self) -> AsyncGenerator[AsyncSession, None]:
        session = None
        try:
            async with self.async_session_maker() as session:
                yield session
        finally:
            if session:
                await session.close()



from datetime import datetime

from pydantic import BaseModel


class Credentails(BaseModel):
    access_token: str
    refresh_token: str
    login: str
    password: str
    access_until: datetime
    service_name: str
    created_at: datetime



import codecs

import gostcrypto


class Kuznechik:
    def __init__(self, key: str):
        self.key = bytearray.fromhex(key)

    def encrypt(self, text: str) -> str:
        heximal = text.encode().hex()
        return self.kuznechik_encryptor(bytes.fromhex(heximal)).hex()

    def kuznechik_encryptor(self, input_array: bytes) -> bytes:
        cipher_obj = gostcrypto.gostcipher.new(
            "kuznechik", self.key, gostcrypto.gostcipher.MODE_ECB, pad_mode=gostcrypto.gostcipher.PAD_MODE_1
        )
        return cipher_obj.encrypt(bytearray(input_array))  # pyright: ignore[reportAttributeAccessIssue, reportReturnType]

    def decrypt(self, text: str) -> str:
        result = self.kuznechik_decryptor(bytes.fromhex(text))
        return codecs.decode(result.hex(), "hex").decode()

    def kuznechik_decryptor(self, input_array: bytes) -> bytes:
        cipher_obj = gostcrypto.gostcipher.new(
            "kuznechik", self.key, gostcrypto.gostcipher.MODE_ECB, pad_mode=gostcrypto.gostcipher.PAD_MODE_1
        )
        result = cipher_obj.decrypt(bytearray(input_array))  # pyright: ignore[reportAttributeAccessIssue]
        return self.remove_fillers(result)  # pyright: ignore[reportArgumentType]

    def remove_fillers(self, byte_array: bytes) -> bytes:
        return byte_array.replace(b"\x00", b"")


class Streebog:
    def __init__(self, key: str):
        self.key = bytearray.fromhex(key)

    def encrypt(self, text: str) -> str:
        heximal = text.encode().hex()
        return self.streebog_encryptor(bytes.fromhex(heximal)).hex()

    def streebog_encryptor(self, input_array: bytes) -> bytearray:
        cipher_obj = gostcrypto.gosthash.new("streebog512", data=input_array)
        return cipher_obj.digest()
